"""Le sous-module scrippy_core.workspace apporte les objets nécessaires à la gestion de l'espace de travail."""
import os
import shutil
import pathlib
import logging
import scrippy_core
from scrippy_core.error_handler import ScrippyCoreError


class Workspace():
  """La classe Workspace offre un espace de travail dont le répertoire est basé sur le le nom passé en argument du constructeur.

  L'espace de travail sera alors créé dans le répoetoire scrippy_core.SCRIPPY_TMPDIR.
  """

  def __init__(self, name):
    """Initialise l'espace de travail."""
    self.path = pathlib.Path(os.path.join(scrippy_core.SCRIPPY_TMPDIR, name))

  def __enter__(self):
    """Point d'entrée."""
    logging.info(f"[+] Creation de l'espace de travail: {self.path}")
    try:
      self.path.mkdir(mode=0o750, parents=True, exist_ok=False)
    except Exception as err:
      err_msg = f"Erreur lors de la creation de l'espace de travail temporaire: [{err.__class__.__name__}] {err}"
      logging.critical(err_msg)
      raise ScrippyCoreError(err_msg) from err
    return str(self.path)

  def __exit__(self, kind, value, traceback):
    """Point de sortie."""
    logging.info(f"[+] Suppression de l'espace de travail: {self.path}")
    try:
      shutil.rmtree(self.path)
    except Exception as err:
      err_msg = f"Erreur lors de la suppression de l'espace de travail temporaire: [{err.__class__.__name__}] {err}"
      logging.critical(err_msg)
      raise ScrippyCoreError(err_msg) from err

  def __eq__(self, other):
    """Surcharge l'implémentation par défaut."""
    if isinstance(other, Workspace):
      return self.path == other.path
    return False

  def __str__(self):
    """Renvoie une représentation sous forme de chaîne de caractères."""
    return str(self.path)



"""
Ce module propose une interface à la configuration des scripts basés sur le cadriciel Scrippy.

Cette interface permet:
- Le chargement et la prise en compte automatique d'un fichier de configuration
- Le controle de la validité de la configuration
"""

import os
import logging
import configparser
import scrippy_core
from scrippy_core.context import Context
from scrippy_core.scriptinfo import ScriptInfo
from scrippy_core.error_handler import ScrippyCoreError


class Config:
  """
  Config est la classe principale du module scrippy_core.conf.

  Elle permet le chargement automatique de la configuration et, si la *docstring* du script comporte un ensemble de
  ligne commençant par `@conf`, permet le contrôle de validité de la configuration du fichier de configuration.

  Les lignes des déclarations du format de configuration doivent respecter le formalisme suivant:

  @conf:<section>|<clé>|<type_valeur>

  <type_valeur> doit être l'un des types reconnus suivants:
  - str (chîne de caractères)
  - int (entier)
  - float (nombre à virgule flotante)
  - bool (booléen)

  Exemple:
  --------
  À partir de la déclaration suivante:

  @conf:log/level/str
  @conf:database/port/int
  @conf:sql/verbose/bool

  Le fichier de configuration suivant sera vérifié:
  [log]
    level = ERROR
  [database]
    port = 5432
  [sql]
    verbose = True
  """

  def __init__(self):
    """Initialise l'instance.

    Aucun contrôle des valeurs des paramètres n'est effectué.
    Seule la structure de la configuration (présence des sections et clés décrites par l'ensemble `@conf`) est vérifiée. Un contrôle du type des clés déclarées est également effectué.
    """
    self.conf = configparser.ConfigParser()
    context = Context.get_current()
    if context is not None:
      script_name = context.name
      self.doc = context.doc
    else:
      script_name = ScriptInfo.get_script_name()
      self.doc = ScriptInfo.get_doc()
    self.config_filename = os.path.join(scrippy_core.SCRIPPY_CONFDIR,
                                        f"{script_name}.conf")
    if os.path.isfile(self.config_filename):
      logging.info(f"[+] Chargement du fichier de configuration {self.config_filename}")
      self.conf.read(self.config_filename)
    else:
      logging.info(f"[+] Pas de fichier de configuration {self.config_filename}")
    self._check()

  def _check(self):
    """
    Contrôle de la configuration.

    Cette méthode est appelée automatiquement par le constructeur si l'argument optionnel 'params' est fourni.
    Cette méthode n'a pas à être appelée autrement que par le constructeur de la classe Config
    """
    context = Context.get_current()
    if context is None:
      context = Context.get_root_context()
    logging.info("[+] Controle de la configuration")
    conf_lines = []
    conf_lines = [line.split(':')[1].strip() for line in self.doc.splitlines() if line.strip().startswith("@conf")]
    for line in conf_lines:
      try:
        section, key, param_type, secret = line.split("|")
        value = self.get(section, key, param_type)
        if secret.lower() == "true":
          context.vault.add(value)
      except ValueError:
        section, key, param_type = line.split("|")
        value = self.get(section, key, param_type)
      logging.debug(f"[{section}][{key}] = {value} [OK]")
    return True

  def register_secrets(self, context):
    logging.info("[+] Enregistrement des secrets (conf)")
    conf_lines = []
    conf_lines = [line.split(':')[1].strip() for line in self.doc.splitlines() if line.strip().startswith("@conf")]
    for line in conf_lines:
      try:
        section, key, param_type, secret = line.split("|")
        value = self.get(section, key, param_type)
        if secret.lower() == "true":
          context.vault.add(value)
      except ValueError:
        pass

  def get_section(self, section):
    """
    Renvoie l'ensemble des paires clef/valeur de la section passée en argument.

    Lève une  si la section demandée n'existe pas.
    """
    sec = {}
    try:
      for key in self.conf[section]:
        sec[key] = self.conf[section][key]
      return sec
    except configparser.NoSectionError as err:
      err_msg = f"Section inconnue: [{section}]"
      logging.critical(err_msg)
      raise ScrippyCoreError(err_msg) from err

  def get(self, section, key, param_type='str'):
    """
    Renvoie la valeur d'une clef dont le nom et la section sont passées en argument.

    À moins que le paramètre 'param_type' soit positionné à l'une des valeurs autorisées (str (défaut), int, float ou bool), le type renvoyé est systématiquement une chaine de caractère.
    """
    try:
      return {'int': self.conf.getint,
              'float': self.conf.getfloat,
              'bool': self.conf.getboolean,
              'str': self.conf.get}[param_type](section, key)
    except configparser.NoOptionError as err:
      err_msg = f"Parametre inconnu: [{section}][{key}]"
      logging.critical(err_msg)
      raise ScrippyCoreError(err_msg) from err
    except configparser.NoSectionError as err:
      err_msg = f"Section inconnue: [{section}]"
      logging.critical(err_msg)
      raise ScrippyCoreError(err_msg) from err
    except ValueError as err:
      err_msg = f"Erreur de type: [{section}][{key}] n'est pas de type '{param_type}'"
      logging.critical(err_msg)
      raise ScrippyCoreError(err_msg) from err
    except KeyError as err:
      err_msg = f"Erreur d'usage: Type de parametre inexistant: '{param_type}'"
      logging.critical(err_msg)
      raise ScrippyCoreError(err_msg) from err

  def has(self, section, key):
    """Renvoie True si la clé section.key existe dans la config."""
    return self.conf.has_option(section, key)



"""Gestionnaire d'erreurs."""
import sys
import traceback
import logging


class ScrippyCoreError(Exception):
  """Classe d'erreur spécifique."""

  def __init__(self, message):
    """Initialise l'instance."""
    self.message = message
    super().__init__(self.message)


def handle_error(kind, value, tb):
  """Récupère toutes les erreurs et sort proprement après avoir logguer les erreurs rencontrées."""
  if kind is not None:
    logging.critical(f"[{kind.__name__}]: {str(value)}")
    if logging.getLogger().getEffectiveLevel() <= logging.DEBUG:
      traceback.print_tb(tb)
    if kind != SystemExit:
      sys.exit(1)



#!/usr/bin/python3
"""
--------------------------------------------------------------------------------
  @author         : Michaël Costa (michael.costa@mcos.nc)
  @date           : 04/02/2021
  @version        : 1.0.0
  @description    : Script de test

--------------------------------------------------------------------------------
  Mise a jour :
  1.0.0  04/02/2021 - Michaël Costa - Cre: Création

--------------------------------------------------------------------------------
  Liste des utilisateurs ou groupe autorisés:
    @user:root

--------------------------------------------------------------------------------
  Liste des paramètres de configuration obligatoires:
    <section>|<clef>|<type>|<secret>
    @conf:database|host|str
    @conf:database|port|int
    @conf:database|user|str|true
    @conf:database|password|str|true
    @conf:database|db|str
    @conf:git|host|str
    @conf:git|port|int
    @conf:git|user|str|true
    @conf:git|repo|str
    @conf:git|branch|str
    @conf:git|ssh_cmd|str
    @conf:smtp|host|str
    @conf:smtp|port|int
    @conf:pop|host|str
    @conf:pop|port|int
    @conf:pop|user|str|true
    @conf:pop|password|str|true
    @conf:ssh|host|str
    @conf:ssh|port|int
    @conf:ssh|user|str|true
    @conf:ssh|keyfile|str|true
    @conf:api|host|str
    @conf:api|port|int
    @conf:api|endpoint|str


--------------------------------------------------------------------------------
  Liste des paramètres des options et arguments d'exécution:
    <nom>|<type>|<help>|<num_args>|<defaut>|<conflit>|<implique>|<requis>|<secret>
    @args:remote_path|str|Chemin distant|1||||true|false
    @args:remote_filename|str|nom du fichier distant|1||||true|false
    @args:local_path|str|Chemin local|1||||true|false
    @args:pattern|str|Motif|1|.*|||false|false
    @args:mail_recipient|choice|Adresse de courriel du destinataire|1|harry.fink@flying.circus,luiggi,vercotti@flying.circus|||true|true


--------------------------------------------------------------------------------
  Fonctionnement:
  ---------------
    Script de test du cadriciel Scrippy.

    Ce script exploite un panel des fonctionnalités proposées par le cadriciel Scrippy telles que:
    - Exploitation d'API
    - Manipulation de base de données
    - Envoi (SMTP) et reception (POP3) de courriel
    - Exécution de commandes distantes (SSH)
    - Transfert de fichiers (SFTP)
    - Manipulation de dépôts Git
    - Manipulation de fichiers modèles (Template Jinja2)

  D'autres fonctionnalités sont disponibles: https://codeberg.org/scrippy
"""
# ------------------------------------------------------------------------------
#  Initialisation de l’environnement
# ------------------------------------------------------------------------------
import os
import json
import email
import shutil
import logging
import scrippy_core
from scrippy_db.db import Db
from scrippy_git.git import Repo
from scrippy_remote.remote import Ssh
from scrippy_api.api.client import Client
from scrippy_template.template import Renderer
from scrippy_mail.mail import Mailer, PopClient


# ------------------------------------------------------------------------------
#  fonctions
# ------------------------------------------------------------------------------
def get_api_user_list(host, port, endpoint):
  """Récupère la liste des utilisateurs à partir de l'API."""
  logging.info("[+] Récupération de la liste des utilisateurs depuis l'API")
  params = {}
  url = f"http://{host}:{port}/{endpoint}"
  client = Client(verify=False)
  response = client.request(method="GET", url=url, params=params)
  assert response.status_code == 200
  users = response.json()['users']
  for user in users:
    logging.info(f" '-> First name: {user['first_name']} / Last name: {user['last_name']} / Password: {user['password']}")
  return response.json()


def add_api_user(host, port, endpoint, user):
  """Ajoute un utilisateur à l'aide de l'API."""
  logging.info("[+] Ajout de l'utilisateur à l'aide de l'API")
  for attr in user["user"]:
    logging.info(f" '-> {attr}: {user['user'][attr]}")
  url = f"http://{host}:{port}/{endpoint}"
  client = Client(verify=False)
  response = client.request(method="PUT", url=url, data=json.dumps(user))
  assert response.status_code == 200
  return response.json()


def get_db_users(username, host, port, database, password):
  """Récupère la liste des utilisateurs à partir de la base de données."""
  logging.info("[+] Récupération de la liste des utilisateurs depuis la base de données")
  with Db(username=username, host=host, port=port, database=database, password=password) as db:
    req = "select id, name, givenname, password from users order by id;"
    params = None
    users = db.execute(req, params)
    for user in users:
      logging.info(f" '-> ID: {user[0]} / Last name: {user[1]} / First name: { user[2]} / Password: {user[3]}")
    return users


def add_db_user(username, host, port, database, password, user):
  """Ajout une utilisateur dans la base de données."""
  logging.info("[+] Ajout d'un utilisateur dans la base de données")
  for attr in user:
    logging.info(f" '-> {attr}")
  with Db(username=username, host=host, port=port, database=database, password=password) as db:
    req = "insert into users values (DEFAULT, %s, %s, %s) returning id;"
    params = user
    result = db.execute(req, params=params, commit=True)
    logging.info(result)
    return result


def render_template(template, params):
  """Rendu de fichiers modèles."""
  logging.info("[+] rendu du fichier modèle")
  logging.info(f" '-> {template}")
  base_path = scrippy_core.SCRIPPY_TEMPLATEDIR
  renderer = Renderer(base_path, template)
  return renderer.render(params=params)


def send_mail(host, port, rcpt, body):
  """Envoi d'un courriel."""
  logging.info("[+] Envoi du courriel")
  mailer = Mailer(host=host, port=port, starttls=False)
  mail_from = "luiggi.vercotti@flying.circus"
  mail_to = (rcpt, )
  mail_subject = "Nobody expects the spanish inquisition"
  mail_body = body
  return mailer.send(mail_subject, mail_body, mail_to, mail_from)


def get_mail(host, port, username, password):
  """Récupération d'un courriel."""
  logging.info("[+] Récupération du courriel")
  pop_client = PopClient(host=host, port=port)
  pop_client.connect()
  pop_client.authenticate(username, password)
  num_available_mails = pop_client.stat().decode()[4]
  logging.info(f" '-> Nombre de courriels disponibles: {num_available_mails}")
  assert num_available_mails == "1"
  logging.info("  '-> Ok")
  mail_content = pop_client.retr(1)
  mail_content = email.message_from_bytes(mail_content).get_payload()
  mail_content = mail_content.replace("=\r\n", "")
  mail_content = mail_content.replace("\r\n.\r\n", "\n")
  mail_content = "\n".join(mail_content.split("\r\n"))
  logging.info(f" '-> {mail_content}")
  return mail_content


def delete_mail(host, port, username, password, num):
  """Suppression des courriels."""
  logging.info("[+] Suppression des courriels")
  pop_client = PopClient(host=host, port=port)
  pop_client.connect()
  pop_client.authenticate(username, password)
  return pop_client.dele(num)


def git_clone(repo, branch, path, env):
  """Clonage de dépôt Git."""
  logging.info("[+] Clonage du dépôt:")
  repo.clone(branch=branch, path=path, env=env)


def git_push(repo, message):
  """Pousse les modifications vers le dépôt distant."""
  logging.info("[+] Envoi des modifications")
  repo.commit_push(message)


def ssh_send_files(username, hostname, port, key_filename, local_path, remote_path, pattern):
  """Envoi de fichiers sur hôte distant."""
  logging.info("[+] Envoi des fichiers via SSH")
  with Ssh(username=username, hostname=hostname, port=port, key_filename=key_filename) as remote_host:
    local_files = remote_host.find_local_files(local_path, pattern, recursive=True)
    err = remote_host.transfer_local_files(local_files, remote_path, exit_on_error=True)
    err += remote_host.delete_local_files(local_files, exit_on_error=True)
    return err


def ssh_exec_cmd(username, hostname, port, key_filename, cmd):
  """Exécution d'une commande sur hôte distant."""
  logging.info("[+] Exécution de la commande sur l'hôte distant")
  logging.info(f" '-> {cmd}")
  with Ssh(username=username, hostname=hostname, port=port, key_filename=key_filename) as remote_host:
    return remote_host.exec_command(cmd, return_stdout=True)


def ssh_get_file(username, hostname, port, key_filename, local_path, remote_path, remote_filename):
  """Récupération de fichier depuis l'hôte distant."""
  pattern = os.path.join(remote_path, remote_filename)
  logging.info("[+] Récupération du fichier distant")
  logging.info(f" '-> {pattern}")
  with Ssh(username=username, hostname=hostname, port=port, key_filename=key_filename) as remote_host:
    remote_files = remote_host.find_remote_files(remote_path, pattern, recursive=True, exit_on_error=True)
    err = remote_host.transfer_remote_files(local_path, remote_files, exit_on_error=True)
    err += remote_host.delete_remote_files(remote_files, exit_on_error=True)
    return err


# ------------------------------------------------------------------------------
#  Traitement principal
# ------------------------------------------------------------------------------
def main(args):
  with scrippy_core.ScriptContext(__file__) as _context:
    config = _context.config
    logging.info("[+] Test started...")
    # ------------------------------------------------------------------------------
    #  Test du client d'API
    # ------------------------------------------------------------------------------
    new_user = {"user": {"first_name": "Harry", "last_name": "Fink", "password": "D3ADP4RR0T"}}
    expected = {"user": new_user["user"], "method": "PUT"}
    result = add_api_user(config.get("api", "host"), config.get("api", "port"), config.get("api", "endpoint"), new_user)
    assert result == expected
    result = get_api_user_list(config.get("api", "host"), config.get("api", "port"), config.get("api", "endpoint"))
    assert len(result["users"]) == 11
    assert new_user["user"] in result["users"]
    # ------------------------------------------------------------------------------
    #  Test du client Base de données
    # ------------------------------------------------------------------------------
    user = ("VERCOTTI", "Luiggi", "SP4N15H1NQU1S1T10N")
    result = add_db_user(config.get("database", "user"),
                         config.get("database", "host"),
                         config.get("database", "port", "int"),
                         config.get("database", "db"),
                         config.get("database", "password"),
                         user)
    assert result[0][0] == 1
    users = get_db_users(config.get("database", "user"),
                         config.get("database", "host"),
                         config.get("database", "port", "int"),
                         config.get("database", "db"),
                         config.get("database", "password"))
    assert len(users) == 2 and users[0][1] == "FINK" and users[0][2] == "Harry"
    assert users[1][0] == 1, users[1][1] == "VERCOTTI" and users[1][2] == "Luiggi"
    # ------------------------------------------------------------------------------
    #  Test du moteur de modèles
    # ------------------------------------------------------------------------------
    params = {"user": {"first_name": users[0][2],
                       "last_name": users[0][1],
                       "password": users[0][3]},
              "sender": {"first_name": users[1][2],
                         "last_name": users[1][1]}}
    mail_body = render_template(template=config.get("templates", "filename"), params=params)
    logging.info(mail_body)
    with open(os.path.join(scrippy_core.SCRIPPY_TMPDIR, "message.txt"), "w") as message:
      message.write(mail_body)
    # ------------------------------------------------------------------------------
    #  Test du client SMTP
    # ------------------------------------------------------------------------------
    assert send_mail(host=config.get("smtp", "host"),
                     port=config.get("smtp", "port", "int"),
                     rcpt=args.mail_recipient,
                     body=mail_body)
    # ------------------------------------------------------------------------------
    #  Test du client POP3
    # ------------------------------------------------------------------------------
    mail_content = get_mail(host=config.get("pop", "host"),
                            port=config.get("pop", "port", "int"),
                            username=config.get("pop", "user"),
                            password=config.get("pop", "password"))
    assert mail_content.strip() == mail_body.strip()
    dele_result = delete_mail(host=config.get("pop", "host"),
                              port=config.get("pop", "port", "int"),
                              username=config.get("pop", "user"),
                              password=config.get("pop", "password"),
                              num=1)
    logging.info(dele_result)
    assert dele_result.decode()[:3] == "+OK"
    # ------------------------------------------------------------------------------
    #  Test du client Git
    # ------------------------------------------------------------------------------
    ssh_cmd = config.get("git", "ssh_cmd").format(os.path.dirname(os.path.realpath(__file__)))
    local_path = os.path.join(scrippy_core.SCRIPPY_TMPDIR, config.get("git", "repo"))
    ENV = {"GIT_SSH_VARIANT": "ssh",
           "GIT_SSH_COMMAND": ssh_cmd}
    repo = Repo(username=config.get("git", "user"),
                host=config.get("git", "host"),
                port=config.get("git", "port", "int"),
                reponame=config.get("git", "repo"))
    git_clone(repo, branch=config.get("git", "branch"), path=local_path, env=ENV)
    shutil.copy(os.path.join(scrippy_core.SCRIPPY_TMPDIR, "message.txt"),
                os.path.join(local_path, "message.txt"))
    git_push(repo, "Harry fink is a new user")
    # ------------------------------------------------------------------------------
    #  Test du client SSH
    # ------------------------------------------------------------------------------
    key_filename = config.get("ssh", "keyfile").format(os.path.dirname(os.path.realpath(__file__)))
    err = ssh_send_files(username=config.get("ssh", "user"),
                         hostname=config.get("ssh", "host"),
                         port=config.get("ssh", "port", "int"),
                         key_filename=key_filename,
                         local_path=args.local_path,
                         remote_path=args.remote_path,
                         pattern=args.pattern)
    assert err == 0
    cmd = f"find {args.remote_path} -type f"
    stdout = ssh_exec_cmd(username=config.get("ssh", "user"),
                          hostname=config.get("ssh", "host"),
                          port=config.get("ssh", "port", "int"),
                          key_filename=key_filename,
                          cmd=cmd)
    assert stdout["exit_code"] == 0
    err = ssh_get_file(username=config.get("ssh", "user"),
                       hostname=config.get("ssh", "host"),
                       port=config.get("ssh", "port", "int"),
                       key_filename=key_filename,
                       local_path=os.path.dirname(args.local_path),
                       remote_path=args.remote_path,
                       remote_filename=args.remote_filename)
    assert err == 0
    assert os.path.isfile(os.path.join(os.path.dirname(args.local_path), args.remote_filename))

    # ------------------------------------------------------------------------------
    logging.info("[+] Test finished !")


# ------------------------------------------------------------------------------
#  Point d'entrée
# ------------------------------------------------------------------------------
if __name__ == '__main__':
  main(scrippy_core.args)



"""Le module scrippy_core.context propose l'ensemble des objets nécessaire à la gestion du contexte d'exécution des scripts."""
import os
import re
import time
import errno
import logging
import inspect
import importlib
from filelock import FileLock
from scrippy_core.scriptinfo import ScriptInfo


class Vault:
  """La classe Vault gère les informations sensibles qui ne devraient pas apparaître dans les logs."""

  def __init__(self):
    self.secrets = []
    self.pattern = None

  def add(self, value):
    self.secrets.append(str(value))
    self.pattern = re.compile(f"{'|'.join(self.secrets)}")

  def is_secret(self, value):
    return value in self.secrets

  def protect(self, msg):
    if self.pattern is not None:
      return re.sub(self.pattern, "*******", str(msg))
    return msg


class PIDStack:
  """La classe PIDStack permet la gestion des exécution concurrentielles."""

  def __init__(self, filename, tmpdir):
    """Initialise la pile d'exécution."""
    filename = os.path.basename(filename)
    filename = os.path.splitext(filename)[0]
    self.stack_file_path = os.path.join(tmpdir, f"{filename}.pids")
    self.lock = FileLock(f"{self.stack_file_path}.lock")

  def _read(self):
    pids = []
    if os.path.isfile(self.stack_file_path):
      with open(self.stack_file_path, mode="r") as stack_file:
        pids = [pid.strip() for pid in stack_file.readlines()]
      if pids is None:
        pids = []
    return pids

  def _write(self, pids):
    with open(self.stack_file_path, mode="w") as stack_file:
      for pid in pids:
        stack_file.write(f"{pid}\n")

  def get_pids(self):
    """Renvoie la liste des PID en attente ou en cours d'exécution."""
    pids = []
    with self.lock.acquire():
      for pid in self._read():
        try:
          os.kill(int(pid), 0)
          pids.append(int(pid))
        except OSError as e:
          if e.errno == errno.EPERM:
            pids.append(pid)
          elif e.errno == errno.ESRCH:
            pass
      self._write(pids)
      return pids

  def register(self):
    """Enregistre un nouveau PID dans la pile."""
    with self.lock.acquire():
      pids = self.get_pids()
      pids.append(os.getpid())
      self._write(pids)

  def checkout(self):
    """Fais le ménage dans la pile."""
    with self.lock.acquire():
      self._write([pid for pid in self.get_pids() if pid != str(os.getpid())])


class Context:
  """La classe Context permet la gestion du contexte d'exécution du script."""

  _root_context = None

  def __init__(self, filename, session, root=False, temp=False):
    """Initialise le contexte d'exécution."""
    self.filename = os.path.basename(filename)
    self.name = os.path.splitext(self.filename)[0]
    self.session = session
    self.vault = Vault()
    self.childs = []
    self.root = root
    self.temp = temp
    logging.getLogger('filelock').setLevel(logging.CRITICAL)
    self.log_session_name = f"{self.name}_{self.session}"
    if root:
      self.doc = ScriptInfo.get_doc()
    else:
      self.doc = importlib.import_module(self.name).__doc__

  @staticmethod
  def create(filename):
    """Crée le contacte."""
    Context.get_root_context()  # Force la creation du rootContext
    if Context._root_context.temp:
      new_context = Context(filename,
                            session=Context._root_context.session,
                            root=True)
      Context._root_context = new_context
    else:
      session = f"{time.time()}_{os.getpid()}"
      new_context = Context(filename, session=session)
    parent_context = Context.get_current()
    if parent_context:
      Context.find(parent_context)._add_child(new_context)
    return new_context

  @staticmethod
  def find(context_to_find, current_context=None):
    """Retrouve le contexte d'exécution cpmplet."""
    if not current_context:
      current_context = Context._root_context
    if current_context == context_to_find:
      return current_context
    else:
      for child in current_context.childs:
        found = Context.find(context_to_find, child)
        if found:
          return found
      return None

  @staticmethod
  def get_root_context():
    """Renvoie le contexte racine."""
    if not Context._root_context:
      session = f"{time.time()}_{os.getpid()}"
      Context._root_context = Context(ScriptInfo.get_script_filename(),
                                      session=session,
                                      root=True,
                                      temp=True)
    return Context._root_context

  @staticmethod
  def get_current():
    """Renvoie le contexte courant."""
    for fi in inspect.stack():
      if '_context' in fi[0].f_locals and isinstance(fi[0].f_locals['_context'], Context):
        context = fi[0].f_locals['_context']
        return context
    return None

  @staticmethod
  def print_full_stack(context=None, lvl=0):
    """Affiche la pile complète des contextes."""
    if not context:
      context = Context.get_root_context()
    print(' |-' + lvl * '--' + context.name + ' ' + context.session)
    for child in context.childs:
      Context.print_full_stack(child, lvl + 1)

  def _add_child(self, child_context):
    self.childs.append(child_context)

  def print_stack(self):
    """Affiche la pile."""
    lvl = 0
    for context in self.stack():
      print(' |-' + lvl * '--' + context.name)
      lvl += 1

  def get_parent(self, tree=None):
    tree = tree or Context.get_root_context()
    for context in tree.childs:
      if context == self:
        return tree
      else:
        res = self.get_parent(context)
        if res:
          return res
    return None

  def stack(self):
    stack = []
    current = self
    while current:
      stack.append(current)
      current = current.get_parent()
    stack.reverse()
    return stack

  def __eq__(self, other):
    """Overrides the default implementation"""
    if isinstance(other, Context):
      return self.name == other.name and self.session == other.session
    return False

  def __str__(self):
    return f"Context(name='{self.name}', filename='{self.filename}', session='{self.session}', childs='{list(map(lambda child: str(child), self.childs))}', root={self.root})"



"""Le sous-module scrippy_core.scriptinfo."""
import os
import __main__


class ScriptInfo:
  """L'objet ScriptInfo."""

  @staticmethod
  def get_script_full_filename():
    """Retourne le chemin complet du script courant."""
    return __main__.__file__

  @staticmethod
  def get_script_filename():
    """Retourne le nom de fichier du script courant."""
    return os.path.basename(__main__.__file__)

  @staticmethod
  def get_script_name():
    """Retourne le nom du script courant."""
    return os.path.splitext(ScriptInfo.get_script_filename())[0]

  @staticmethod
  def get_doc():
    """Retourne la documentation du script courant."""
    return __main__.__doc__



"""Le sous-module scrippy_core.log fournit les objets nécessaires à la gestion des fichiers de journalisation."""
import os
import sys
import logging
import coloredlogs
import scrippy_core
from scrippy_core.context import Context
from scrippy_core.log.infralogger import InfraLogger


def setLoggerClass(klass):
  """https://stackoverflow.com/questions/51938027/python-3-7-logging-setloggerclass-isnt ."""
  logging.getLogger().__class__ = klass
  logging.setLoggerClass(klass)


class LogConfig:
  """Configurateur du logger."""

  console_fmt = '[%(asctime)s] [%(levelname)-8s] %(message)s'
  file_fmt = '[%(asctime)s] [%(levelname)-8s] [%(name)s] %(message)s'
  datefmt = '%Y%m%d %H:%M:%S'
  field_styles = {'asctime': {'color': 'magenta'},
                  'hostname': {'color': 'magenta'},
                  'levelname': {'color': 'white', 'bold': True},
                  'programname': {'color': 'cyan'},
                  'name': {'color': 'cyan'}}
  level_styles = {'debug': {'color': 'white'},
                  'info': {'color': 'green'},
                  'warning': {'color': 'yellow'},
                  'error': {'color': 'red'},
                  'critical': {'color': 'red', 'bold': True}}

  @staticmethod
  def default_configuration():
    """Met en place la configuration par défaut du logger."""
    console_handler = logging.StreamHandler(sys.stdout)
    # TERM vaut "dumb" dans JobScheduler
    if "TERM" in os.environ and os.environ["TERM"] != "dumb":
      console_handler.setFormatter(coloredlogs.ColoredFormatter(fmt=LogConfig.console_fmt,
                                                                datefmt=LogConfig.datefmt,
                                                                level_styles=LogConfig.level_styles,
                                                                field_styles=LogConfig.field_styles))
    else:
      console_handler.setFormatter(logging.Formatter(fmt=LogConfig.console_fmt,
                                                     datefmt=LogConfig.datefmt))
    logging.basicConfig(level=logging.INFO,
                        handlers=[console_handler])

  @staticmethod
  def add_file_handler():
    """Ajoute un fichier journal."""
    context = Context.get_root_context()
    log_filename = LogConfig.get_filename_for_session(context.session)
    file_handler = logging.FileHandler(log_filename)
    file_handler.setFormatter(coloredlogs.ColoredFormatter(fmt=LogConfig.file_fmt,
                                                           datefmt=LogConfig.datefmt,
                                                           level_styles=LogConfig.level_styles,
                                                           field_styles=LogConfig.field_styles))

    logging.getLogger().addHandler(file_handler)

  @staticmethod
  def get_filename_for_session(session):
    """Renvoie le nom du fichier de journalisation du script courant."""
    context = Context.get_root_context()
    log_filename = f"{context.name}_{session}.log"
    return os.path.join(scrippy_core.SCRIPPY_LOGDIR, log_filename)

  @staticmethod
  def render_logfile(session):
    """Affiche le journal de la session demandée."""
    if session is not None:
      logfile = LogConfig.get_filename_for_session(session)
      with open(logfile, 'r') as log:
        print(log.read())


setLoggerClass(InfraLogger)



"""Scrippy Core module."""


import os
import grp
import pwd
import time
import yaml
import logging
from scrippy_core.conf import Config
from scrippy_core import error_handler
from scrippy_core.history import History
from scrippy_core.arguments import Parser
from scrippy_core.workspace import Workspace
from scrippy_core.context import Context, PIDStack
from scrippy_core.log.debuglogger import DebugLogger
from scrippy_core.log import LogConfig, setLoggerClass
from scrippy_core.error_handler import ScrippyCoreError


def check_users(context):
  """
  Vérifie que l'utilisateur courant est autorisé à exécuter le script en comparant les déclarations "@user" dans le cartouche du script.

  Lève une ScrippyCoreError si l'utilisateur n'est pas autorisé.
  """
  user_id = os.geteuid()
  try:
    users = [line.split(':')[1].strip() for line in context.doc.splitlines() if line.strip().startswith("@user")]
    users = [pwd.getpwnam(user).pw_uid for user in users]
    if len(users) > 0 and user_id not in users:
      raise ScrippyCoreError('[BadUserError] Utilisateur non autorisé')
  except KeyError as erro:
    raise ScrippyCoreError(f"[UnknownUserError] Utilisateur inconnu: {str(erro)}") from erro
  except Exception as erro:
    err_msg = f"[{erro.__class__.__name__}] {erro}"
    logging.critical(err_msg)
    raise ScrippyCoreError(err_msg) from erro


def check_groups(context):
  """
  Vérifie que l'utilisateur courant est autorisé à exécuter le script en comparant les déclarations "@group" dans le cartouche du script.

  Lève une ScrippyCoreError si l'utilisateur ne fait pas partie des groupes autorisés.
  """
  user_groups = os.getgroups()
  try:
    groups = [line.split(':')[1].strip() for line in context.doc.splitlines() if line.strip().startswith("@group")]
    groups = [grp.getgrnam(group)[2] for group in groups]
    if len(groups) > 0:
      if not len([groups for g in user_groups if g in groups]) > 0:
        raise ScrippyCoreError("[BadGroupError] L'utilisateur ne fait pas partie des groupes autorisés")
  except KeyError as erro:
    raise ScrippyCoreError(f"[UnknownGroupError] Groupe inconnu: {str(erro)}") from erro
  except Exception as erro:
    err_msg = f"[{erro.__class__.__name__}] {erro}"
    logging.critical(err_msg)
    raise ScrippyCoreError(err_msg) from erro


def check_instances(context):
  """
  Vérifie que le nombre d'instances autorisées à être simultanément exécutées n'est pas atteint en comparant la déclaration @max_instance dans le cartouche du script et le nombr de PID retourné par la pidstack.

  Si le nombre d'instance est atteint, le script fera une pause jusqu'à ce que le nombre d'instance en cours d'exécution soit inférieur au nombre d'instances autorisées.
  Lorsque le nombre maximum d'instance autorisées n'est pas atetint, le script s'enregistre dans la pidstack et s'exécute.
  """
  sleep_step = 3
  bools = ["true", "1", "on"]
  try:
    max_instance = [line.split(':')[1].strip() for line in context.doc.splitlines() if line.strip().startswith("@max_instance")][0]
  except IndexError:
    max_instance = 0
  try:
    timeout = [line.split(':')[1].strip() for line in context.doc.splitlines() if line.strip().startswith("@timeout")][0]
  except IndexError:
    timeout = 0
  try:
    exit_on_wait = [line.split(':')[1].strip() for line in context.doc.splitlines() if line.strip().startswith("@exit_on_wait")][0]
  except IndexError:
    exit_on_wait = "False"
  try:
    exit_on_timeout = [line.split(':')[1].strip() for line in context.doc.splitlines() if line.strip().startswith("@exit_on_timeout")][0]
  except IndexError:
    exit_on_timeout = "False"
  pids = context.pidstack.get_pids()
  try:
    timeout = int(timeout)
    max_instance = int(max_instance)
    exit_on_timeout = exit_on_timeout.lower() in bools
    exit_on_wait = exit_on_wait.lower() in bools
    if max_instance > 0 and len(pids) > max_instance:
      logging.info(f"[+] En attente d'un creneau d'execution: {len(pids)}/{max_instance} [{timeout}s]")
      while len(pids) > max_instance and pids[0] != os.getpid():
        timeout -= sleep_step
        if timeout <= 0 and exit_on_timeout:
          raise Exception("TimeoutError: Delai d'attente depasse")
        if exit_on_wait:
          raise Exception(f"EagernessError: `exit_on_wait` positionne a {exit_on_wait}")
        pids = context.pidstack.get_pids()
        time.sleep(sleep_step)
  except Exception as erro:
    err_msg = f"[{erro.__class__.__name__}] {erro}"
    logging.critical(err_msg)
    raise ScrippyCoreError(err_msg) from erro


# ------------------------------------------------------------------------------
# INITIALISATION
# ------------------------------------------------------------------------------
conf_files = ["/etc/scrippy/scrippy.yml",
              os.path.expanduser("~/.config/scrippy/scrippy.yml"),
              "/usr/local/etc/scrippy/scrippy.yml"]

for conf_file in conf_files:
  if os.path.isfile(conf_file):
    with open(conf_file, mode="r", encoding="utf-8") as conf_file:
      scrippy_conf = yaml.load(conf_file, Loader=yaml.FullLoader)
      SCRIPPY_LOGDIR = scrippy_conf.get("env").get("logdir")
      SCRIPPY_HISTDIR = scrippy_conf.get("env").get("histdir")
      SCRIPPY_REPORTDIR = scrippy_conf.get("env").get("reportdir")
      SCRIPPY_TMPDIR = scrippy_conf.get("env").get("tmpdir")
      SCRIPPY_DATADIR = scrippy_conf.get("env").get("datadir")
      SCRIPPY_TEMPLATEDIR = scrippy_conf.get("env").get("templatedir")
      SCRIPPY_CONFDIR = scrippy_conf.get("env").get("confdir")

conf_keys = [SCRIPPY_LOGDIR,
             SCRIPPY_HISTDIR,
             SCRIPPY_REPORTDIR,
             SCRIPPY_TMPDIR,
             SCRIPPY_DATADIR,
             SCRIPPY_TEMPLATEDIR,
             SCRIPPY_CONFDIR]

for conf_key in conf_keys:
  if conf_key is None:
    raise ScrippyCoreError(f"Clef de configuration manquante: {conf_key}")

LogConfig.default_configuration()
arg_parser = Parser()
args = arg_parser.args
_log_file_enabled = True

if args.nolog:
  logging.getLogger().setLevel(logging.ERROR)

if args.debug:
  logging.warning("[+] Option --debug activee -> le niveau de log ne peut plus etre modifie")
  logging.getLogger().setLevel(logging.DEBUG)
  setLoggerClass(DebugLogger)

logging.debug(f"[+] Arguments : {vars(args)}")

if args.no_log_file:
  _log_file_enabled = False
  logging.warning("[+] Option --no-log-file activee -> pas de fichier de log, ni d'historique pour cette execution")

# Chargement et contrôle du fichier de configuration root
root_config = Config()

if root_config.has('log', 'level'):
  logging.info(f"[!] Configuration de la journalisation: {root_config.get('log', 'level').upper()}")
  logging.getLogger().setLevel(root_config.get('log', 'level').upper())

if root_config.has('log', 'file') and \
   not root_config.get('log', 'file', 'bool'):
  logging.warning("[+] Config log.file = False -> pas de fichier de log, ni d'historique pour cette execution")
  _log_file_enabled = False

if _log_file_enabled:
  LogConfig.add_file_handler()


class ScriptContext:
  """Contexte d'exécution du script."""

  def __init__(self, name, retention=50, workspace=False):
    """Initalisation du contexte."""
    self.name = name
    self.hist_enabled = True
    self.worskspace_enabled = workspace
    # La variable _context est nécessaire à getCurrentContext()
    # Touche pas à ça p'tit con©
    _context = Context.create(self.name)
    self.context = _context

    if self.context.root:
      self.context.config = root_config
    else:
      self.context.config = Config()
    self.context.config.register_secrets(self.context)
    arg_parser.register_secrets(self.context)

    if self.worskspace_enabled:
      self.workspace = Workspace(self.context.log_session_name)

    if _log_file_enabled:
      self.context.hist = History(retention=retention)
    else:
      self.hist_enabled = False

    check_users(self.context)
    check_groups(self.context)
    try:
      self.context.pidstack = PIDStack(self.name, SCRIPPY_TMPDIR)
      self.context.pidstack.register()
    except PermissionError as erro:
      err_msg = f"Erreur lors de la création de la pile des instances: [{erro.__class__.__name__}] {erro}"
      logging.critical(err_msg)
      raise ScrippyCoreError(err_msg) from erro
    check_instances(self.context)

  def __enter__(self):
    """Point d'entrée."""
    # La variable _context est nécessaire à getCurrentContext()
    # Touche pas à ça p'tit con©
    _context = self.context
    if self.hist_enabled:
      self.context.hist.__enter__()
    if self.worskspace_enabled:
      self.context.workspace_path = self.workspace.__enter__()
    return self.context

  def __exit__(self, kind, value, tb):
    """Point de sortie."""
    # La variable _context est nécessaire à getCurrentContext()
    # Touche pas à ça p'tit con©
    _context = self.context
    if self.worskspace_enabled:
      self.workspace.__exit__(kind, value, tb)
    if self.hist_enabled:
      self.context.hist.__exit__(kind, value, tb)
    self.context.pidstack.checkout()
    error_handler.handle_error(kind, value, tb)



"""Le sous-module scrippy_core.log.debuglogger permet la journalisation en mode déverminage."""
import logging
from scrippy_core.log.infralogger import InfraLogger


class DebugLogger(InfraLogger):
  """Gestionnaire de journaux spécifique au mode déverminage."""

  def __init__(self, name, level=logging.NOTSET):
    """Initialise le gestionnaire de journaux en mode déverminage."""
    super(DebugLogger, self).__init__(name, level)

  def setLevel(self, level):
    """Surcharge logging.Logger.setLevel."""
    # On ne fait rien le niveau de log est forcé à DEBUG
    pass



"""Le sous-module scrippy_core.log.debuglogger permet la journalisation en mode normal."""
import logging
from scrippy_core.context import Context


class InfraLogger(logging.Logger):
  """Gestionnaire de journaux."""

  def __init__(self, name, level=logging.NOTSET):
    """Initialise le gestionnaire de journaux en mode normal."""
    super(InfraLogger, self).__init__(name, level)

  def makeRecord(self, name, level, fn, lno, msg, args, exc_info, func=None, extra=None, sinfo=None):
    """Surcharge logging.Logger.makeRecord."""
    script_name = name
    context = Context.get_current()
    if context is None:
      context = Context.get_root_context()
    script_name = context.log_session_name
    msg = context.vault.protect(msg)
    return super(InfraLogger, self).makeRecord(script_name, level, fn, lno, msg, args, exc_info, func, extra, sinfo)
